# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions related to scattering matrix computation for the FMM algorithm.

Copyright (c) Martin F. Schubert
"""

import dataclasses
import functools
from typing import Sequence, Tuple

import jax
import jax.numpy as jnp
from jax import tree_util

from fmmax import fmm, misc, utils
from fmmax._scattering import (  # noqa: F401
    ScatteringMatrix,
    _extend_s_matrix,
    append_layer,
    prepend_layer,
    set_end_layer_thickness,
    set_start_layer_thickness,
)


def stack_s_matrix(
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    force_x64_solve: bool = False,
) -> "ScatteringMatrix":
    """Computes the s-matrix for a stack of layers.

    If only a single layer is provided, the scattering matrix is just the
    identity matrix, and start and end layer data is for the same layer.

    Args:
        layer_solve_results: The eigensolve results for layers in the stack.
        layer_thicknesses: The scalar thicknesses for layers in the stack.
        force_x64_solve: If ``True``, matrix solves will be done with 64 bit precision.

    Returns:
        The ``ScatteringMatrix``.
    """
    return _stack_s_matrices(
        layer_solve_results, layer_thicknesses, force_x64_solve=force_x64_solve
    )[-1]


def stack_s_matrices_interior(
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    force_x64_solve: bool = False,
) -> Tuple[Tuple["ScatteringMatrix", "ScatteringMatrix"], ...]:
    """Computes scattering matrices before and after each layer in the stack.

    Specifically, for each layer ``i`` two ``ScatteringMatrix`` are returned. The
    first relates fields in the substack ``0, ..., i``, while the second relates
    the fields in the substack ``i, ..., N``, where ``N`` is the maximum layer
    index. These two scattering matrices are denoted as the corresponding
    to the "before" substack and the "after" substack.

    Args:
        layer_solve_results: The eigensolve results for layers in the stack.
        layer_thicknesses: The scalar thicknesses for layers in the stack.
        force_x64_solve: If ``True``, matrix solves will be done with 64 bit precision.

    Returns:
        The tuple of ``(scattering_matrix_before, scattering_matrix_after)``.
    """
    before = _stack_s_matrices(
        layer_solve_results,
        layer_thicknesses,
        force_x64_solve=force_x64_solve,
    )

    # Compute the scattering matrix for the substack "after" each layer. We do
    # this by computing the scattering matrix for the substack "before" each layer
    # in the reversed stack. Then, reverse each resulting scattering matrix.
    reverse = _stack_s_matrices(
        layer_solve_results[::-1],
        layer_thicknesses[::-1],
        force_x64_solve=force_x64_solve,
    )
    after = tuple(
        [
            ScatteringMatrix(
                s11=s.s22,
                s12=s.s21,
                s21=s.s12,
                s22=s.s11,
                start_layer_solve_result=s.end_layer_solve_result,
                start_layer_thickness=s.end_layer_thickness,
                end_layer_solve_result=s.start_layer_solve_result,
                end_layer_thickness=s.start_layer_thickness,
            )
            for s in reverse[::-1]
        ]
    )
    return tuple(zip(before, after))


def _stack_s_matrices(
    layer_solve_results: Sequence[fmm.LayerSolveResult],
    layer_thicknesses: Sequence[jnp.ndarray],
    force_x64_solve: bool,
) -> Tuple["ScatteringMatrix", ...]:
    """Computes the s-matrices for a stack of layers.

    The matrices are for various substacks: the first scattering matrix is
    for the first layer only, the second is for the stack consisting of the
    first and second layer, etc.

    Note that the ``LayerSolveResult`` attributes of the resulting ``ScatteringMatrix``
    has the ``tangent_vector_field`` attribute stripped out.

    Args:
        layer_solve_results: The eigensolve results for layers in the stack. All
            eigensolve results must have identical batch shape.
        layer_thicknesses: The scalar thicknesses for layers in the stack.
        force_x64_solve: If ``True``, matrix solves will be done with 64 bit precision.

    Returns:
        The tuple of ``ScatteringMatrix``.
    """
    if len(layer_solve_results) != len(layer_thicknesses):
        raise ValueError(
            f"`layer_solve_results` and `layer_thicknesses` should have the same "
            f"length but got {len(layer_solve_results)} and {len(layer_thicknesses)}."
        )

    # Remove the tangent vector fields from the solve results.
    layer_solve_results = tuple(
        dataclasses.replace(solve_result, tangent_vector_field=None)
        for solve_result in layer_solve_results
    )

    # Broadcast all layer solve results so their attributes have the full batch shape.
    batch_shape = jnp.broadcast_shapes(
        *[lsr.batch_shape for lsr in layer_solve_results]
    )
    layer_solve_results = [
        fmm.broadcast_result(lsr, batch_shape) for lsr in layer_solve_results
    ]

    # Broadcast all layer thicknesses so they have a common shape.
    t_shape = jnp.broadcast_shapes(*[jnp.shape(t) for t in layer_thicknesses])
    layer_thicknesses = [jnp.broadcast_to(t, t_shape) for t in layer_thicknesses]

    # The initial scattering matrix is just the identity matrix, with the
    # necessary batch dimensions.
    eye = misc.diag(jnp.ones_like(layer_solve_results[0].eigenvalues))
    layer_s_matrix_0 = ScatteringMatrix(
        s11=eye,
        s12=jnp.zeros_like(eye),
        s21=jnp.zeros_like(eye),
        s22=eye,
        start_layer_solve_result=layer_solve_results[0],
        start_layer_thickness=layer_thicknesses[0],
        end_layer_solve_result=layer_solve_results[0],
        end_layer_thickness=layer_thicknesses[0],
    )

    if len(layer_solve_results) == 1:
        return (layer_s_matrix_0,)

    layer_s_matrix_1 = _pair_s_matrix(
        layer_solve_result=layer_solve_results[0],
        layer_thickness=layer_thicknesses[0],
        next_layer_solve_result=layer_solve_results[1],
        next_layer_thickness=layer_thicknesses[1],
        force_x64_solve=force_x64_solve,
    )

    if len(layer_solve_results) == 2:
        return (layer_s_matrix_0, layer_s_matrix_1)

    # If we have more than two layers, stack the remaining layer solve results and
    # thicknesses so they can be scanned over.

    stacked_remaining_solve_results = tree_util.tree_map(
        lambda *xs: jnp.stack(xs, axis=0),
        *layer_solve_results[2:],
    )
    stacked_remaining_thicknesses = jnp.stack(layer_thicknesses[2:])

    def scan_fn(s_matrix, args):
        next_layer_solve_result, next_layer_thickness = args
        s_matrix = append_layer(
            s_matrix,
            next_layer_solve_result,
            next_layer_thickness,
            force_x64_solve=force_x64_solve,
        )
        return s_matrix, s_matrix

    _, stacked_remaining_s_matrices = jax.lax.scan(
        scan_fn,
        init=layer_s_matrix_1,
        xs=(stacked_remaining_solve_results, stacked_remaining_thicknesses),
    )

    # Unstack to return a tuple of scattering matrices.
    remaining_s_matrices = tuple(
        tree_util.tree_map(lambda x: x[i, ...], stacked_remaining_s_matrices)
        for i in range(len(layer_solve_results) - 2)
    )
    return (layer_s_matrix_0, layer_s_matrix_1) + remaining_s_matrices


def stack_s_matrix_scan(
    layer_solve_results: fmm.LayerSolveResult,
    layer_thicknesses: jnp.ndarray,
    force_x64_solve: bool = False,
) -> "ScatteringMatrix":
    """Computes the scattering matrix for a stack of layers by a scan operation.

    Unlike ``stack_s_matrix``, this function uses a scan operation rather than a python
    for loop, which can lead to significantly smaller programs and shorter compile
    times. However, it requires that the layer solve results and thicknesses for be
    represented in the leading batch dimension of the arguments.

    This function is best used when the eigensolve for each layer is also done in a
    batched manner.

    Args:
        layer_solve_results: The layer solve results for all layers in the stack.
        layer_thicknesses: The scalar layer thicknesses for all layers in the stack.
        force_x64_solve: If ``True``, matrix solves will be done with 64 bit precision.

    Returns:
        The scattering matrix for the stack.
    """
    assert layer_thicknesses.ndim == 1
    if layer_solve_results.batch_shape[0] != len(layer_thicknesses):
        raise ValueError(
            f"`layer_solve_results` and `layer_thicknesses` should be compatible (i.e. "
            f"correspond to the same number of layers) but inferred layer numbers of "
            f"{layer_solve_results.batch_shape[0]} and {layer_thicknesses.shape[0]}."
        )

    eye = misc.diag(
        jnp.ones(
            layer_solve_results.eigenvalues.shape[1:],
            dtype=layer_solve_results.eigenvalues.dtype,
        )
    )
    start_solve_result = tree_util.tree_map(lambda x: x[0, ...], layer_solve_results)
    s_matrix = ScatteringMatrix(
        s11=eye,
        s12=jnp.zeros_like(eye),
        s21=jnp.zeros_like(eye),
        s22=eye,
        start_layer_solve_result=start_solve_result,
        start_layer_thickness=layer_thicknesses[0],
        end_layer_solve_result=start_solve_result,
        end_layer_thickness=layer_thicknesses[0],
    )

    def scan_fn(s_matrix, x):
        next_layer_solve_result, next_layer_thickness = x
        s_matrix = append_layer(
            s_matrix,
            next_layer_solve_result,
            next_layer_thickness,
            force_x64_solve=force_x64_solve,
        )
        return s_matrix, s_matrix

    s_matrix, _ = jax.lax.scan(
        scan_fn,
        init=s_matrix,
        xs=(
            tree_util.tree_map(lambda x: x[1:], layer_solve_results),
            layer_thicknesses[1:],
        ),
    )
    return s_matrix


def redheffer_star_product(
    a: "ScatteringMatrix",
    b: "ScatteringMatrix",
    force_x64_solve: bool = False,
) -> "ScatteringMatrix":
    """Compute the Redheffer star product of two scattering matrices."""
    a_extended = append_layer(a, b.start_layer_solve_result, b.start_layer_thickness)
    a11, a12, a21, a22 = a_extended.s11, a_extended.s12, a_extended.s21, a_extended.s22
    b11, b12, b21, b22 = b.s11, b.s12, b.s21, b.s22

    solve = functools.partial(utils.solve, force_x64_solve=force_x64_solve)

    # See https://en.wikipedia.org/wiki/Redheffer_star_product
    eye = misc.diag(jnp.ones_like(a11[..., 0]))
    s11 = b11 @ solve(eye - a12 @ b21, a11)
    s12 = b12 + b11 @ solve(eye - a12 @ b21, a12 @ b22)
    s21 = a21 + a22 @ solve(eye - b21 @ a12, b21 @ a11)
    s22 = a22 @ solve(eye - b21 @ a12, b22)
    return ScatteringMatrix(
        s11=s11,
        s12=s12,
        s21=s21,
        s22=s22,
        start_layer_solve_result=a.start_layer_solve_result,
        start_layer_thickness=a.start_layer_thickness,
        end_layer_solve_result=b.end_layer_solve_result,
        end_layer_thickness=b.end_layer_thickness,
    )


def _pair_s_matrix(
    layer_solve_result: fmm.LayerSolveResult,
    layer_thickness: jnp.ndarray,
    next_layer_solve_result: fmm.LayerSolveResult,
    next_layer_thickness: jnp.ndarray,
    force_x64_solve: bool,
) -> "ScatteringMatrix":
    """Generate the scattering matrix for a pair of layers."""
    # Alias for brevity: eigenvalues, eigenvectors, and omega-k matrix.
    q = layer_solve_result.eigenvalues
    phi = layer_solve_result.eigenvectors
    omega_k = layer_solve_result.omega_script_k_matrix

    next_q = next_layer_solve_result.eigenvalues
    next_phi = next_layer_solve_result.eigenvectors
    next_omega_k = next_layer_solve_result.omega_script_k_matrix

    solve = functools.partial(utils.solve, force_x64_solve=force_x64_solve)

    term1 = q[..., jnp.newaxis] * solve(
        omega_k @ phi,
        next_omega_k @ next_phi * (1 / next_q)[..., jnp.newaxis, :],
    )
    term2 = solve(omega_k @ phi, omega_k @ next_phi)
    i11 = i22 = 0.5 * (term1 + term2)
    i12 = i21 = 0.5 * (-term1 + term2)

    fd = jnp.exp(1j * q * layer_thickness)
    fd_next = jnp.exp(1j * next_q * next_layer_thickness)

    # The computation is identical to that in `_extend_s_matrix` with `s11` and `s22`
    # being the identity, and `s12` and `s21` being zero.
    fd_diag = jnp.expand_dims(misc.diag(fd), tuple(range(i11.ndim - fd.ndim - 1)))
    s11 = solve(i11, fd_diag)
    s12 = solve(i11, -i12 * fd_next[..., jnp.newaxis, :])
    s21 = i21 @ s11
    s22 = i21 @ s12 + i22 * fd_next[..., jnp.newaxis, :]

    return ScatteringMatrix(
        s11=s11,
        s12=s12,
        s21=s21,
        s22=s22,
        start_layer_solve_result=layer_solve_result,
        start_layer_thickness=layer_thickness,
        end_layer_solve_result=next_layer_solve_result,
        end_layer_thickness=next_layer_thickness,
    )



"""Functions related to layer eigenmode calculation for the FMM algorithm.

Copyright (c) 2023 Meta Platforms, Inc. and affiliates.
"""

import dataclasses
from typing import Optional, Tuple

import jax.numpy as jnp
import numpy as onp

from fmmax import basis, misc


@dataclasses.dataclass
class LayerSolveResult:
    """Stores the result of a layer eigensolve.

    This eigenvalue problem is specified in equation 28 of [2012 Liu].

    Attributes:
        wavelength: The wavelength for the solve.
        in_plane_wavevector: The in-plane wavevector for the solve.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.
        expansion: The expansion used for the eigensolve.
        eigenvalues: The layer eigenvalues.
        eigenvectors: The layer eigenvectors.
        omega_script_k_matrix: The omega-script-k matrix from equation 26 of [2012 Liu].
        z_permittivity_matrix: The fourier-transformed zz-component of permittivity.
        inverse_z_permittivity_matrix: The fourier-transformed inverse of zz-component
            of permittivity.
        transverse_permittivity_matrix: The transverse permittivity matrix which relates
            the electric field and electric displacement fields.
        z_permeability_matrix: The fourier-transformed zz-component of permeability.
        inverse_z_permeability_matrix: The fourier-transformed inverse of zz-component
            of permeability.
        transverse_permeability_matrix: The transverse permeability matrix, needed to
            calculate the omega-script-k matrix from equation 26 of [2012 Liu]. This
            is needed to generate the layer scattering matrix.
        tangent_vector_field: The tangent vector field ``(tx, ty)`` used to compute the
            transverse permittivity matrix, if a vector FMM formulation is used. If
            the ``FFT`` formulation is used, the vector field is ``None``.
    """

    wavelength: jnp.ndarray
    in_plane_wavevector: jnp.ndarray
    primitive_lattice_vectors: basis.LatticeVectors
    expansion: basis.Expansion
    eigenvalues: jnp.ndarray
    eigenvectors: jnp.ndarray
    omega_script_k_matrix: jnp.ndarray
    z_permittivity_matrix: jnp.ndarray
    inverse_z_permittivity_matrix: jnp.ndarray
    transverse_permittivity_matrix: jnp.ndarray
    z_permeability_matrix: jnp.ndarray
    inverse_z_permeability_matrix: jnp.ndarray
    transverse_permeability_matrix: jnp.ndarray
    tangent_vector_field: Optional[Tuple[jnp.ndarray, jnp.ndarray]]

    @property
    def batch_shape(self) -> Tuple[int, ...]:
        return self.eigenvectors.shape[:-2]

    def __post_init__(self) -> None:
        """Validates shapes of the ``LayerSolveResult`` attributes."""
        # Avoid validation when attributes are e.g. tracers.
        if not isinstance(self.eigenvalues, (jnp.ndarray, onp.ndarray)):
            return

        required_dtype = self.eigenvalues.dtype
        if self.eigenvectors.dtype != required_dtype:
            raise ValueError(
                f"`eigenvectors` should have dtype {required_dtype} but got "
                f"{self.eigenvectors.dtype}"
            )
        if self.z_permittivity_matrix.dtype != required_dtype:
            raise ValueError(
                f"`z_permittivity_matrix` should have dtype {required_dtype} but got "
                f"{self.z_permittivity_matrix.dtype}"
            )

        if self.inverse_z_permittivity_matrix.dtype != required_dtype:
            raise ValueError(
                f"`inverse_z_permittivity_matrix` should have dtype {required_dtype} "
                f"but got {self.inverse_z_permittivity_matrix.dtype}"
            )
        if self.z_permeability_matrix.dtype != required_dtype:
            raise ValueError(
                f"`z_permeability_matrix` should have dtype {required_dtype} but got "
                f"{self.z_permeability_matrix.dtype}"
            )
        if self.transverse_permeability_matrix.dtype != required_dtype:
            raise ValueError(
                f"`transverse_permeability_matrix` should have dtype {required_dtype} "
                f"but got {self.transverse_permeability_matrix.dtype}"
            )

        def _incompatible(arr: jnp.ndarray, reference_shape: Tuple[int, ...]) -> bool:
            ndim_mismatch = arr.ndim != len(reference_shape)
            batch_compatible = misc.batch_compatible_shapes(arr.shape, reference_shape)
            return ndim_mismatch or not batch_compatible

        if _incompatible(self.wavelength, self.batch_shape):
            raise ValueError(
                f"`wavelength` must have compatible batch shape, but got shape "
                f"{self.wavelength.shape} when `eigenvectors` shape is "
                f"{self.eigenvectors.shape}."
            )
        if _incompatible(self.in_plane_wavevector, self.batch_shape + (2,)):
            raise ValueError(
                f"`in_plane_wavevector` must have compatible batch shape, but got "
                f"shape {self.in_plane_wavevector.shape} when `eigenvectors` shape is "
                f"{self.eigenvectors.shape}."
            )
        if _incompatible(self.primitive_lattice_vectors.u, self.batch_shape + (2,)):
            raise ValueError(
                f"`primitive_lattice_vectors.u` must have compatible batch shape, but "
                f"got shape {self.primitive_lattice_vectors.u.shape} when "
                f"`eigenvectors` shape is {self.eigenvectors.shape}."
            )
        if _incompatible(self.primitive_lattice_vectors.v, self.batch_shape + (2,)):
            raise ValueError(
                f"`primitive_lattice_vectors.v` must have compatible batch shape, but "
                f"got shape {self.primitive_lattice_vectors.v.shape} when "
                f"`eigenvectors` shape is {self.eigenvectors.shape}."
            )
        if self.expansion.num_terms * 2 != self.eigenvectors.shape[-1]:
            raise ValueError(
                f"`eigenvectors` must have shape compatible with `expansion.num_terms`,"
                f" but got shape {self.eigenvectors.shape} when `num_terms` shape is "
                f"{self.expansion.num_terms}."
            )
        if self.eigenvalues.shape != self.eigenvectors.shape[:-1]:
            raise ValueError(
                f"`eigenvalues` must have compatible shape, but got shape "
                f"{self.eigenvalues.shape} when `eigenvectors` shape is "
                f"{self.eigenvectors.shape}."
            )
        if self.omega_script_k_matrix.shape != self.eigenvectors.shape:
            raise ValueError(
                f"`omega_script_k_matrix` must have shape matching `eigenvectors` "
                f"shape, but got {self.eigenvalues.shape} when `eigenvectors` shape "
                f"is {self.eigenvectors.shape}."
            )

        expected_matrix_shape = self.batch_shape + (self.expansion.num_terms,) * 2
        if _incompatible(self.inverse_z_permittivity_matrix, expected_matrix_shape):
            raise ValueError(
                f"`inverse_z_permittivity_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.inverse_z_permittivity_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.z_permittivity_matrix, expected_matrix_shape):
            raise ValueError(
                f"`z_permittivity_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes {self.z_permittivity_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.transverse_permittivity_matrix, self.eigenvectors.shape):
            raise ValueError(
                f"`transverse_permittivity_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.transverse_permittivity_matrix.shape} and "
                f"{self.eigenvectors.shape}."
            )
        if _incompatible(self.inverse_z_permeability_matrix, expected_matrix_shape):
            raise ValueError(
                f"`inverse_z_permeability_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.inverse_z_permeability_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.z_permeability_matrix, expected_matrix_shape):
            raise ValueError(
                f"`z_permeability_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes {self.z_permeability_matrix.shape} "
                f"and {self.eigenvectors.shape}."
            )
        if _incompatible(self.transverse_permeability_matrix, self.eigenvectors.shape):
            raise ValueError(
                f"`transverse_permeability_matrix` must have shape compatible with "
                f"`eigenvectors`, but got shapes "
                f"{self.transverse_permeability_matrix.shape} and "
                f"{self.eigenvectors.shape}."
            )

        if self.tangent_vector_field is not None and (
            self.tangent_vector_field[0].ndim != self.eigenvectors.ndim
        ):
            raise ValueError(
                f"`tangent_vector_field` must have ndim compatible with "
                f"`eigenvectors`, but got shapes {self.tangent_vector_field[0]} and "
                f"{self.eigenvectors}."
            )



"""Functions related transforming to and from the Fourier basis.

Copyright (c) 2023 Meta Platforms, Inc. and affiliates.
"""

import jax.numpy as jnp

from fmmax import basis


def _standard_toeplitz_indices(expansion: basis.Expansion) -> jnp.ndarray:
    """Computes the indices for a standard Toeplitz matrix for `basis_coefficients`.

    Args:
        expansion: The field expansion to be used.

    Returns:
        The indices, with shape `(num, num, 2)`.
    """
    i, j = jnp.meshgrid(
        jnp.arange(expansion.num_terms),
        jnp.arange(expansion.num_terms),
        indexing="ij",
    )
    basis_coefficients = jnp.asarray(expansion.basis_coefficients)
    idx = basis_coefficients[i, :] - basis_coefficients[j, :]
    return idx



"""Defines functions related to farfield patterns in the FMM scheme.

These functions are intended for calculations that involve Brillouin
zone integration, and batched in-plane wavevectors as generated by
`basis.brillouin_zone_in_plane_wavevector`.

Copyright (c) 2023 Meta Platforms, Inc. and affiliates.
"""

from typing import Callable, Tuple

import jax
import jax.numpy as jnp
import numpy as onp

from fmmax import basis, misc, utils


def farfield_profile(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
) -> Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]:
    """Computes a farfield profile.

    This function effectively "unstacks" the values for each Fourier order and
    for each point in the Brillouin zone sampling scheme.

    Args:
        flux: The flux array, with shape ``(..., num_bz_kx, num_bz_ky, ...,
            2 * num_terms, num_sources)``.
        wavelength: The wavelength, batch-compatible with `flux`.
        in_plane_wavevector: The in-plane wavevector for the zeroth Fourier
            order, batch-compatible with `flux`.
        primitive_lattice_vectors: The primitive lattice vectors of the unit cell.
        expansion: The expansion used for the fields.
        brillouin_grid_axes: Specifies the two axes of ``flux`` corresponding to
            the Brillouin zone grid.

    Returns:
        The polar and azimuthal angles, solid angle associated with each value,
        and the farfield power.
    """
    assert flux.shape[-2] == 2 * expansion.num_terms
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    ndim_batch = flux.ndim - 2
    wavelength = misc.atleast_nd(wavelength, ndim_batch)
    in_plane_wavevector = misc.atleast_nd(in_plane_wavevector, ndim_batch + 1)

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )
    transverse_wavevectors = _unflatten_transverse_wavevectors(
        transverse_wavevectors, expansion, brillouin_grid_axes
    )
    flux = _unflatten_flux(flux, expansion, brillouin_grid_axes)

    # Remove the brillouin zone axes from `wavelength`, making it compatible
    # with the unflattened flux and transverse wavevectors.
    wavelength = jnp.squeeze(wavelength, axis=brillouin_grid_axes)

    polar_angle, azimuthal_angle = _angles_from_unflattened_transverse_wavevectors(
        transverse_wavevectors=transverse_wavevectors,
        wavelength=wavelength,
    )

    # Transform flux form units of power per unit Brillouin zone area to
    # power per unit solid angle. Add dummy dimensions for the polarization
    # and sources.
    solid_angle = _solid_angle_from_unflattened_transverse_wavevectors(
        transverse_wavevectors=transverse_wavevectors,
        wavelength=wavelength,
    )
    transformed_flux = flux / solid_angle[..., jnp.newaxis, jnp.newaxis]
    return polar_angle, azimuthal_angle, solid_angle, transformed_flux


def _angles_from_unflattened_transverse_wavevectors(
    transverse_wavevectors: jnp.ndarray,
    wavelength: jnp.ndarray,
) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Computes the propagation angles in free space for given wavevectors.

    Evanescent modes whose transverse wavevector magnitude exceeds that of
    the free space wavevector are given a polar angle of ``pi / 2``.

    Args:
        transverse_wavevectors: The unflattened transverse wavectors, with
            shape ``(..., nkx, nky, 2)``.
        wavelength: The free-space wavelength.

    Returns:
        Arrays containing the polar and azimuthal angles.
    """
    assert transverse_wavevectors.ndim - 3 == wavelength.ndim
    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]
    kt = jnp.sqrt(kx**2 + ky**2)

    sin_polar_angle = kt * wavelength[..., jnp.newaxis, jnp.newaxis] / (2 * jnp.pi)
    polar_angle = jnp.where(
        sin_polar_angle > 1, jnp.pi / 2, jnp.arcsin(sin_polar_angle)
    )

    azimuthal_angle = jnp.angle(kx + 1j * ky)
    return polar_angle, azimuthal_angle


def _solid_angle_from_unflattened_transverse_wavevectors(
    transverse_wavevectors: jnp.ndarray,
    wavelength: jnp.ndarray,
) -> jnp.ndarray:
    """Computes the solid angle associated with each transverse wavevector.

    The transverse wavevectors should be unflattened, i.e. the ``-3`` and ``-2``
    axes should correspond to different points in k-space.

    Args:
        transverse_wavevectors: The unflattened transverse wavevectors
        wavelength: The free-space wavelength.

    Returns:
        The solid angle, with the shape matching the leading dimensions of
        `transverse_wavevectors`.
    """
    assert transverse_wavevectors.ndim >= 3

    kx = transverse_wavevectors[..., 0]
    ky = transverse_wavevectors[..., 1]

    # Normalize the wavevectors so they lie on the unit sphere.
    kx /= 2 * jnp.pi / wavelength[..., jnp.newaxis, jnp.newaxis]
    ky /= 2 * jnp.pi / wavelength[..., jnp.newaxis, jnp.newaxis]
    polar_angle = jnp.arcsin(jnp.sqrt(kx**2 + ky**2))

    # Each of our transverse wavevectors lies within a "cell" in the kxky plane.
    # Compute the area of each cell, and then project it onto the unit sphere
    # to get the solid angle associated with each transverse wavevector.
    #
    # First, compute the locations of the cell verteces.
    pad_width = ((0, 0),) * (kx.ndim - 2) + ((1, 1), (1, 1))
    vertex_kx = jnp.pad(kx, pad_width, mode="edge")
    vertex_ky = jnp.pad(ky, pad_width, mode="edge")
    vertex_kx = (
        vertex_kx[..., :-1, :-1]
        + vertex_kx[..., :-1, 1:]
        + vertex_kx[..., 1:, :-1]
        + vertex_kx[..., 1:, 1:]
    ) / 4
    vertex_ky = (
        vertex_ky[..., :-1, :-1]
        + vertex_ky[..., :-1, 1:]
        + vertex_ky[..., 1:, :-1]
        + vertex_ky[..., 1:, 1:]
    ) / 4
    vertex_kt = jnp.stack([vertex_kx, vertex_ky], axis=-1)

    # Find the vectors defining each parallelogramic cell.
    v1 = vertex_kt[..., :-1, 1:, :] - vertex_kt[..., :-1, :-1, :]
    v2 = vertex_kt[..., 1:, :-1, :] - vertex_kt[..., :-1, :-1, :]

    # Find the area of each parallelogramic cell.
    cell_area = jnp.abs(v1[..., 0] * v2[..., 1] - v2[..., 0] * v1[..., 1])

    # Project the area onto the unit sphere.
    projected_area = cell_area / jnp.cos(polar_angle)
    return projected_area


# -----------------------------------------------------------------------------
# Functions for computing the total flux in some angular cone.
# -----------------------------------------------------------------------------


def farfield_integrated_flux(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
    angle_bounds_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    upsample_factor: int,
) -> jnp.ndarray:
    """Computes the flux within the bounds defined by ``angle_bounds_fn``.

    Args:
        flux: The flux array, with shape ``(..., num_bz_kx, num_bz_ky, ...
            2 * num_terms, num_sources)``.
        wavelength: The wavelength, batch-compatible with ``flux``.
        in_plane_wavevector: The in-plane wavevector for the zeroth Fourier
            order, batch-compatible with ``flux``.
        primitive_lattice_vectors: The primitive lattice vectors of the unit cell.
        expansion: The expansion used for the fields.
        brillouin_grid_axes: Specifies the two axes of ``flux`` corresponding to
            the Brillouin zone grid.
        angle_bounds_fn: A function with signature ``fn(polar_angle,
            azimuthal_angle) -> mask`` returning a mask that is ``True`` for angles
            that should be included in the integral.
        upsample_factor: Integer factor specifying upsampling performed in the
            integral, which is used to approximate trapezoidal rule integration.

    Returns:
        The integrated flux, with shape equal to the batch dimensions of flux,
        excluding those for the brillouin zone grid.
    """
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    # Compute the weights array, which reduce the integration weights to
    # an inner product.
    weights = _integrated_flux_weights(
        flux=flux,
        wavelength=wavelength,
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
        brillouin_grid_axes=brillouin_grid_axes,
        angle_bounds_fn=angle_bounds_fn,
        upsample_factor=upsample_factor,
    )

    # Sum over the Brillouin zone and Fourier order axes.
    return jnp.sum(weights * flux, axis=brillouin_grid_axes + (-2,))


def _integrated_flux_weights(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
    angle_bounds_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    upsample_factor: int,
) -> jnp.ndarray:
    """Returns the integration weights for the bounds defined by ``angle_bounds_fn``."""
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    def _integrated_fn(flux):
        assert flux.shape[-1] == 1
        return jnp.sum(
            _integrated_flux_upsampled(
                flux=flux,
                wavelength=wavelength,
                in_plane_wavevector=in_plane_wavevector,
                primitive_lattice_vectors=primitive_lattice_vectors,
                expansion=expansion,
                brillouin_grid_axes=brillouin_grid_axes,
                angle_bounds_fn=angle_bounds_fn,
                upsample_factor=upsample_factor,
            )
        )

    # The weights are just the gradient of the integrated flux with respect
    # to the flat array elements; use a dummy flux having a single source.
    dummy_flux = jnp.ones_like(flux, shape=flux.shape[:-1] + (1,))
    return jax.grad(_integrated_fn)(dummy_flux)


def _integrated_flux_upsampled(
    flux: jnp.ndarray,
    wavelength: jnp.ndarray,
    in_plane_wavevector: jnp.ndarray,
    primitive_lattice_vectors: basis.LatticeVectors,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
    angle_bounds_fn: Callable[[jnp.ndarray, jnp.ndarray], jnp.ndarray],
    upsample_factor: int,
) -> jnp.ndarray:
    """Computes the flux within the bounds defined by ``angle_bounds_fn``."""
    assert upsample_factor >= 1
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    ndim_batch = flux.ndim - 2
    wavelength = misc.atleast_nd(wavelength, ndim_batch)
    in_plane_wavevector = misc.atleast_nd(in_plane_wavevector, ndim_batch + 1)

    transverse_wavevectors = basis.transverse_wavevectors(
        in_plane_wavevector=in_plane_wavevector,
        primitive_lattice_vectors=primitive_lattice_vectors,
        expansion=expansion,
    )

    flux = _unflatten_flux(flux, expansion, brillouin_grid_axes)
    transverse_wavevectors = _unflatten_transverse_wavevectors(
        transverse_wavevectors, expansion, brillouin_grid_axes
    )

    # Remove the `nan`s that are found at array locations having no associated
    # value in the original flattened arrays.
    flux = jnp.where(jnp.isnan(flux), 0, flux)
    transverse_wavevectors = jnp.where(
        jnp.isnan(transverse_wavevectors), 0, transverse_wavevectors
    )

    # Upsample the transverse wavevectors and flux to high resolution.
    assert flux.shape[-2] == 2
    upsampled_flux_shape = flux.shape[:-4] + (
        upsample_factor * flux.shape[-4],  # kx axis
        upsample_factor * flux.shape[-3],  # ky axis
        flux.shape[-2],  # Polarization axis, length 2
        flux.shape[-1],  # Source axis
    )
    flux = jax.image.resize(flux, upsampled_flux_shape, method="linear")

    assert transverse_wavevectors.shape[-1] == 2
    upsampled_wavevector_shape = transverse_wavevectors.shape[:-3] + (
        upsample_factor * transverse_wavevectors.shape[-3],  # kx axis
        upsample_factor * transverse_wavevectors.shape[-2],  # ky axis
        transverse_wavevectors.shape[-1],  # Direction axis.
    )
    transverse_wavevectors = jax.image.resize(
        transverse_wavevectors, upsampled_wavevector_shape, method="linear"
    )

    # Remove the brillouin grid axes from wavelength, and insert axes
    # for the unstacked wavevector.
    wavelength = jnp.squeeze(wavelength, brillouin_grid_axes)

    # Compute polar and azimuthal angles.
    polar_angle, azimuthal_angle = _angles_from_unflattened_transverse_wavevectors(
        transverse_wavevectors, wavelength
    )

    selected = angle_bounds_fn(polar_angle, azimuthal_angle)
    selected = jnp.where(jnp.isnan(selected), False, selected)
    selected = selected[..., jnp.newaxis, jnp.newaxis]

    masked_flux = jnp.where(selected, flux, 0)

    # Sum over the kx, ky, and polarization axes.
    return jnp.sum(masked_flux, axis=(-4, -3, -2)) / upsample_factor**2


# -----------------------------------------------------------------------------
# Functions related to unflattening results of Brillouin zone integration.
# -----------------------------------------------------------------------------


def _unflatten(flat: jnp.ndarray, expansion: basis.Expansion) -> jnp.ndarray:
    """Unflattens an array for a given expansion and Brillouin integration scheme.

    The returned array combines the values associated with all terms in the
    Fourier expansion at all points in the Brillouin zone grid in a single
    array with trailing axes havving shape ``(num_kx, num_ky)``. Elements in the
    output which have no corresponding elements in ``flat`` are given a value
    of ``nan``.

    The flat array should have shape ``(..., num_bz_kx, num_bz_ky, num_terms)``,
    where ``num_terms`` is the number of terms in the Fourier expansion, and the
    ``-3`` and ``-2`` axes are for the Brillouin zone grid, as used e.g. with
    Brillouin zone integration to model localized sources.

    This function assumes that the Brillouin zone is sampled on a regular grid,
    as produced by ``basis.brillouin_zone_in_plane_wavevector``.

    Args:
        flat: The flat array, with shape  ``(..., num_bz_kx, num_bz_ky, num_terms)``.
        expansion: The expansion used for the array.

    Returns:
        The unflattened array, with shape ``(batch_shape, num_kx, num_ky)``.
    """
    assert flat.ndim >= 3
    assert flat.shape[-1] == expansion.num_terms

    i = expansion.basis_coefficients[:, 0]
    j = expansion.basis_coefficients[:, 1]

    batch_shape = flat.shape[:-3]
    bz_grid_shape = flat.shape[-3:-1]

    # The shape of the output array shoudl accomodate all `(i, j)` values.
    shape = batch_shape + (
        (max(i) - min(i) + 1) * bz_grid_shape[0],
        (max(j) - min(j) + 1) * bz_grid_shape[1],
    )

    bz_i, bz_j = onp.meshgrid(
        onp.arange(bz_grid_shape[0]),
        onp.arange(bz_grid_shape[1]),
        indexing="ij",
    )

    offset_i = min(i) * bz_grid_shape[0]
    offset_j = min(j) * bz_grid_shape[1]

    merged_i = i * bz_grid_shape[0] + bz_i[..., onp.newaxis] - offset_i
    merged_j = j * bz_grid_shape[1] + bz_j[..., onp.newaxis] - offset_j

    stacked_i = merged_i.flatten()
    stacked_j = merged_j.flatten()
    stacked_flat = jnp.reshape(flat, batch_shape + (-1,))

    unflattened = jnp.full(shape, jnp.nan, dtype=flat.dtype)
    unflattened = unflattened.at[..., stacked_i, stacked_j].set(stacked_flat)
    return unflattened


def _unflatten_flux(
    flux: jnp.ndarray,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
) -> jnp.ndarray:
    """Unflattens a flux for a given expansion and Brillouin integration scheme.

    Args:
        flux: The flux array, with shape ``(..., num_bz_kx, num_bz_ky, ...
            2 * num_terms, num_sources)``.
        expansion: The expansion used for the flux.
        brillouin_grid_axes: The axes associated with the Brillouin zone grid.

    Returns:
        The unflattened flux, with shape ``(..., num_kx, num_ky, 2, num_sources)``.
    """
    assert flux.ndim >= 4
    assert flux.shape[-2] == 2 * expansion.num_terms
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, flux.ndim
    )

    # The flux array has values for two polarizations at each Fourier order. Split
    # these and treat them as a batch dimension.
    flux = jnp.reshape(flux, flux.shape[:-2] + (2, -1, flux.shape[-1]))

    # Transpose so the axes associated with the Fourier orders and Brillouin zone
    # grid are the trailing axes, as needed by `unflatten`.
    batch_axes = tuple(
        [i for i in range(flux.ndim) if i not in brillouin_grid_axes + (flux.ndim - 2,)]
    )
    axes = batch_axes + tuple([i for i in range(flux.ndim) if i not in batch_axes])
    flux = jnp.transpose(flux, axes)

    flux = _unflatten(flux, expansion)

    # Transpose so that polarization and sources are returned to the trailing axes.
    axes = tuple(range(flux.ndim - 4)) + (-2, -1, -4, -3)
    return jnp.transpose(flux, axes)


def _unflatten_transverse_wavevectors(
    transverse_wavevectors: jnp.ndarray,
    expansion: basis.Expansion,
    brillouin_grid_axes: Tuple[int, int],
) -> jnp.ndarray:
    """Unflatten transverse wavevectors for given expansion and integration scheme.

    Args:
        transverse_wavevectors: The transverse wavevectors array, with shape
            `(..., num_bz_kx, num_bz_ky, ..., num_terms, 2)`.
        expansion: The expansion used for the flux.
        brillouin_grid_axes: The axes associated with the Brillouin zone grid.

    Returns:
        The unflattened wavevectors, with shape `(..., num_kx, num_ky, 2)`.
    """
    assert transverse_wavevectors.ndim >= 4
    assert transverse_wavevectors.shape[-2:] == (expansion.num_terms, 2)
    brillouin_grid_axes = utils.absolute_axes(  # type: ignore[assignment]
        brillouin_grid_axes, transverse_wavevectors.ndim
    )

    # Transpose so the axes associated with the Fourier orders and Brillouin zone
    # grid are the trailing axes, as needed by `unflatten`.
    non_batch_axes = brillouin_grid_axes + (transverse_wavevectors.ndim - 2,)
    batch_axes = tuple(
        [i for i in range(transverse_wavevectors.ndim) if i not in non_batch_axes]
    )
    axes = batch_axes + tuple(
        [i for i in range(transverse_wavevectors.ndim) if i not in batch_axes]
    )
    transverse_wavevectors = jnp.transpose(transverse_wavevectors, axes)

    transverse_wavevectors = _unflatten(transverse_wavevectors, expansion)

    # Transpose so the trailing axis is for the wavevector direction.
    axes = tuple(range(transverse_wavevectors.ndim - 3)) + (-2, -1, -3)
    return jnp.transpose(transverse_wavevectors, axes)



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Defines several utility functions.

Copyright (c) Martin F. Schubert
"""

import jax
import jax.numpy as jnp

from fmmax._utils import (  # noqa: F401
    absolute_axes,
    angular_frequency_for_wavelength,
    interpolate_permittivity,
)


def solve(a: jnp.ndarray, b: jnp.ndarray, *, force_x64_solve: bool) -> jnp.ndarray:
    """Solves ``A @ x = b``, optionally using 64-bit precision."""
    output_dtype = jnp.promote_types(a.dtype, b.dtype)
    if force_x64_solve and jax.config.read("jax_enable_x64"):
        a = a.astype(jnp.promote_types(a.dtype, jnp.float64))
        b = b.astype(jnp.promote_types(b.dtype, jnp.float64))
    return jnp.linalg.solve(a, b).astype(output_dtype)



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions related to layer eigenmode calculation for the FMM algorithm.

Copyright (c) Martin F. Schubert
"""

import enum
from typing import Callable, Tuple

import jax.numpy as jnp

from fmmax import basis, vector
from fmmax._fmm_result import LayerSolveResult

# xx, xy, yx, yy, and zz components of permittivity or permeability.
TensorComponents = Tuple[
    jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray
]
VectorFn = Callable[
    [jnp.ndarray, basis.Expansion, basis.LatticeVectors],
    Tuple[jnp.ndarray, jnp.ndarray],
]


@enum.unique
class Formulation(enum.Enum):
    """Enumerates supported Fourier modal method formulations.

    Each formulation specifies an algorithm to compute the transverse permittivity
    matrix used in the Fourier modal method. The simplest formulation is ``FFT``, in
    which the blocks of the transverse permittivity matrix are simply the Fourier
    convolution matrices for their respective permittivity tensor components.

    The remaining formulations are so-called vector formulations, which make use of a
    vector field generated in the unit cell of the FMM calculation. The vector field
    defines a local coordinate system that tangent and normal to the interfaces of
    features in the unit cell, allowing improved convergence through independent
    treatment of field components that are tangent and normal to the interfaces.
    """

    #: The simplest formulation, which does not consider the orientation of
    #: interfaces of features in a permittivity array.
    FFT = "fft"

    #: Generates a complex linear vector field which has maximum magnitude ``1`` and
    #: a null in the interior of features. In the objective, the gradient of the
    #: vector field on the real-space grid is computed; a penalty term discourages
    #: non-smooth fields (i.e. fields whose gradient is large).
    POL = vector.POL

    #: Takes the field computed by ``POL`` and normalizes so the magnitude is ``1``
    #: evereywhere in the unit cell. Where ``POL`` has zeros, ``NORMAL`` has
    #: discontinuities.
    NORMAL = vector.NORMAL

    #: Takes the field computed by ``POL``, and converts it to a complex
    #: elliptical field which has magnitude ``1`` everywhere and lacks discontinuities.
    JONES = vector.JONES

    #: Directly computes a complex elliptical vector field without first finding
    #: a linear vector field. Smoothness is evaluated on the real-space grid.
    JONES_DIRECT = vector.JONES_DIRECT

    #: Generates a complex linear vector field, but with an alternate method of
    #: penalizing non-smoothness. Specifically, the Fourier components corresponding to
    #: high spatial frequencies are penalized. Compared to ``POL``, ``POL_FOURIER``
    #: can be computed more efficiently.
    POL_FOURIER = vector.POL_FOURIER

    #: Takes the field computed by ``POL_FOURIER``and normalizes so the magnitude is
    #: ``1`` evereywhere in the unit cell.
    NORMAL_FOURIER = vector.NORMAL_FOURIER

    #: Takes the field computed by ``POL_FOURIER`` and converts it to a complex
    #: elliptical field.
    JONES_FOURIER = vector.JONES_FOURIER

    #: Directly computes a complex elliptical vector field, using Fourier coefficients
    #: to penalize non-smoothness.
    JONES_DIRECT_FOURIER = vector.JONES_DIRECT_FOURIER


def broadcast_result(
    layer_solve_result: "LayerSolveResult",
    shape: Tuple[int, ...],
) -> "LayerSolveResult":
    """Broadcast ``layer_solve_result`` attributes to have specified batch shape."""
    lsr = layer_solve_result  # Alias for brevity.
    n = lsr.expansion.num_terms
    return LayerSolveResult(
        wavelength=jnp.broadcast_to(lsr.wavelength, shape),
        in_plane_wavevector=jnp.broadcast_to(lsr.in_plane_wavevector, shape + (2,)),
        primitive_lattice_vectors=basis.LatticeVectors(
            u=jnp.broadcast_to(lsr.primitive_lattice_vectors.u, shape + (2,)),
            v=jnp.broadcast_to(lsr.primitive_lattice_vectors.v, shape + (2,)),
        ),
        expansion=lsr.expansion,
        eigenvalues=jnp.broadcast_to(lsr.eigenvalues, shape + (2 * n,)),
        eigenvectors=jnp.broadcast_to(lsr.eigenvectors, shape + (2 * n, 2 * n)),
        omega_script_k_matrix=jnp.broadcast_to(
            lsr.omega_script_k_matrix, shape + (2 * n, 2 * n)
        ),
        z_permittivity_matrix=jnp.broadcast_to(
            lsr.z_permittivity_matrix, shape + (n, n)
        ),
        inverse_z_permittivity_matrix=jnp.broadcast_to(
            lsr.inverse_z_permittivity_matrix, shape + (n, n)
        ),
        transverse_permittivity_matrix=jnp.broadcast_to(
            lsr.transverse_permittivity_matrix,
            shape + (2 * n, 2 * n),
        ),
        z_permeability_matrix=jnp.broadcast_to(
            lsr.z_permeability_matrix, shape + (n, n)
        ),
        inverse_z_permeability_matrix=jnp.broadcast_to(
            lsr.inverse_z_permeability_matrix, shape + (n, n)
        ),
        transverse_permeability_matrix=jnp.broadcast_to(
            lsr.transverse_permeability_matrix,
            shape + (2 * n, 2 * n),
        ),
        tangent_vector_field=lsr.tangent_vector_field,
    )



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions related to vectors and field expansion in the FMM scheme.

Copyright (c) Martin F. Schubert
"""

from typing import Tuple

import jax.numpy as jnp

from fmmax._basis import (  # noqa: F401
    Expansion,
    LatticeVectors,
    Truncation,
    X,
    Y,
    _basis_coefficients_circular,
    _basis_coefficients_parallelogramic,
    _cross_product,
    _HashableArray,
    _reciprocal,
    generate_expansion,
    min_array_shape_for_expansion,
    plane_wave_in_plane_wavevector,
    transverse_wavevectors,
    unit_cell_coordinates,
    validate_shape_for_expansion,
)


def brillouin_zone_in_plane_wavevector(
    brillouin_grid_shape: Tuple[int, int],
    primitive_lattice_vectors: LatticeVectors,
) -> jnp.ndarray:
    """Compute in-plane wavevectors suitable for Brillouin zone integration.

    The wavevectors are evenly spaced within the first Brillouin zone; for odd grid
    shapes, they subdivide the first Brillouin zone evenly. For even grid shapes, they
    are offset so ``(0, 0)`` is included among the wavevectors.

    Args:
        brillouin_grid_shape: The shape of the wavevector grid.
        primitive_lattice_vectors: The primitive vectors for the real-space lattice.

    Returns:
        The in-plane wavevectors, with shape ``brillouin_grid_shape + (2,)``.
    """
    if len(brillouin_grid_shape) != 2 or brillouin_grid_shape < (1, 1):
        raise ValueError(
            f"`brillouin_grid_shape` must be length-2 with positive values, "
            f"but got {brillouin_grid_shape}."
        )

    udim, vdim = brillouin_grid_shape
    i, j = jnp.meshgrid(
        jnp.arange(-(udim // 2), udim - (udim // 2)) / udim,
        jnp.arange(-(vdim // 2), vdim - (vdim // 2)) / vdim,
        indexing="ij",
    )
    assert i.shape == brillouin_grid_shape
    reciprocal_vectors = primitive_lattice_vectors.reciprocal
    ku = reciprocal_vectors.u
    kv = reciprocal_vectors.v
    return jnp.stack(
        [
            2 * jnp.pi * (i * ku[..., 0] + j * kv[..., 0]),
            2 * jnp.pi * (i * ku[..., 1] + j * kv[..., 1]),
        ],
        axis=-1,
    )



# FMMAX
# Copyright (C) 2025 Martin F. Schubert

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.

# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
"""Functions related to fields in the FMM algorithm.

Copyright (c) Martin F. Schubert
"""

from typing import Tuple

import jax.numpy as jnp

from fmmax._fields import (  # noqa: F401
    _fields_on_grid,
    _layer_fields_3d,
    _stack_fields_3d,
    _validate_amplitudes_shape,
    _validate_matching_lengths,
    amplitude_poynting_flux,
    colocate_amplitudes,
    directional_poynting_flux,
    eigenmode_poynting_flux,
    field_conversion_matrix,
    fields_from_wave_amplitudes,
    fields_on_coordinates,
    fields_on_grid,
    layer_amplitudes_interior,
    layer_fields_3d,
    layer_fields_3d_on_coordinates,
    propagate_amplitude,
    stack_amplitudes_interior,
    stack_amplitudes_interior_with_source,
    stack_fields_3d,
    stack_fields_3d_auto_grid,
    stack_fields_3d_on_coordinates,
)


def time_average_z_poynting_flux(
    electric_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
    magnetic_field: Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray],
) -> jnp.ndarray:
    """Computes the time-average z-directed Poynting flux, given the physical fields.

    The calculation of Poynting flux is an element-wise operation. When the Poynting
    flux is calculated over a single unit cell, the resulting array may be *averaged*
    to yield a flux equal to that in all orders computed by ``amplitude_poynting_flux``.

    Args:
        electric_field: The tuple of electric fields ``(ex, ey, ez)`` defined on the
            real-space grid.
        magnetic_field: The tuple of magnetic fields ``(hx, hy, hz)`` defined on the
            real-space grid.

    Returns:
        The time-average z-directed Poynting flux, with the same shape as ``ex``.
    """
    ex, ey, _ = electric_field
    hx, hy, _ = magnetic_field
    return 0.5 * jnp.real(ex * jnp.conj(hy) - ey * jnp.conj(hx))
